#ifndef HEADER_fdstat_sysctl
#define HEADER_fdstat_sysctl
/*
htop - generic/fdstat_sysctl.h
(C) 2022-2023 htop dev team
Released under the GNU GPLv2+, see the COPYING file
in the source distribution for its full text.
*/


void Generic_getFileDescriptors_sysctl(double* used, double* max);

#endif
