/* Do not edit this file. It was automatically generated. */

#ifndef HEADER_PressureStallMeter
#define HEADER_PressureStallMeter
/*
htop - PressureStallMeter.h
(C) 2004-2011 Hisham H. Muhammad
(C) 2019 Ran Benita
Released under the GNU GPLv2+, see the COPYING file
in the source distribution for its full text.
*/

#include "Meter.h"


extern const MeterClass PressureStallCPUSomeMeter_class;

extern const MeterClass PressureStallIOSomeMeter_class;

extern const MeterClass PressureStallIOFullMeter_class;

extern const MeterClass PressureStallIRQFullMeter_class;

extern const MeterClass PressureStallMemorySomeMeter_class;

extern const MeterClass PressureStallMemoryFullMeter_class;

#endif
