#ifndef HEADER_HugePageMeter
#define HEADER_HugePageMeter
/*
htop - HugePageMeter.h
(C) 2021 htop dev team
Released under the GNU GPLv2+, see the COPYING file
in the source distribution for its full text.
*/

#include "Meter.h"


extern const MeterClass HugePageMeter_class;

#endif /* HEADER_HugePageMeter */
