#ifndef HEADER_UnsupportedProcessTable
#define HEADER_UnsupportedProcessTable
/*
htop - UnsupportedProcessTable.h
(C) 2014 Hisham H. Muhammad
Released under the GNU GPLv2+, see the COPYING file
in the source distribution for its full text.
*/

#include "ProcessTable.h"


typedef struct UnsupportedProcessTable_ {
   ProcessTable super;
} UnsupportedProcessTable;

#endif
